/**
 * Provides classes for working with remote flow sources, sinks and taint propagators
 * from the `github.com/beego/beego` package.
 */

import go
import semmle.go.security.Xss
private import semmle.go.security.SafeUrlFlowCustomizations

// Some TaintTracking::FunctionModel subclasses remain because varargs functions don't work with Models-as-Data sumamries yet.
/**
 * Provides classes for working with remote flow sources, sinks and taint propagators
 * from the [Beego](https://github.com/beego/beego) package.
 */
module Beego {
  /** Gets the v1 module path `github.com/astaxie/beego` or `github.com/beego/beego`. */
  string v1modulePath() { result = ["github.com/astaxie/beego", "github.com/beego/beego"] }

  /** Gets the v2 module path `github.com/beego/beego/v2` */
  string v2modulePath() { result = "github.com/beego/beego/v2" }

  /** Gets the path for the root package of beego. */
  string packagePath() {
    result = package(v1modulePath(), "")
    or
    result = package(v2modulePath(), "server/web")
  }

  /** Gets the path for the context package of beego. */
  string contextPackagePath() {
    result = package(v1modulePath(), "context")
    or
    result = package(v2modulePath(), "server/web/context")
  }

  /** Gets the path for the logs package of beego. */
  string logsPackagePath() {
    result = package(v1modulePath(), "logs")
    or
    result = package(v2modulePath(), "core/logs")
  }

  /** Gets the path for the utils package of beego. */
  string utilsPackagePath() {
    result = package(v1modulePath(), "utils")
    or
    result = package(v2modulePath(), "core/utils")
  }

  /**
   * `BeegoInput` sources of untrusted data.
   */
  private class BeegoInputSource extends RemoteFlowSource::Range {
    string methodName;

    BeegoInputSource() {
      exists(FunctionOutput output |
        methodName = "Bind" and
        output.isParameter(0)
        or
        methodName in [
            "Cookie", "Data", "GetData", "Header", "Param", "Params", "Query", "Refer", "Referer",
            "URI", "URL", "UserAgent"
          ] and
        output.isResult(0)
      |
        exists(DataFlow::MethodCallNode c | this = output.getExitNode(c) |
          c.getTarget().hasQualifiedName(contextPackagePath(), "BeegoInput", methodName)
        )
      )
    }

    predicate isSafeUrlSource() { methodName in ["URI", "URL"] }
  }

  /** `BeegoInput` sources that are safe to use for redirection. */
  private class BeegoInputSafeUrlSource extends SafeUrlFlow::Source {
    BeegoInputSafeUrlSource() { this.(BeegoInputSource).isSafeUrlSource() }
  }

  /**
   * `beego.Controller` sources of untrusted data.
   */
  private class BeegoControllerSource extends RemoteFlowSource::Range {
    BeegoControllerSource() {
      exists(string methodName, FunctionOutput output |
        methodName = "ParseForm" and
        output.isParameter(0)
        or
        methodName in ["GetFile", "GetFiles", "GetString", "GetStrings", "Input"] and
        output.isResult(0)
        or
        methodName = "GetFile" and
        output.isResult(1)
      |
        exists(DataFlow::MethodCallNode c |
          c.getTarget().hasQualifiedName(packagePath(), "Controller", methodName)
        |
          this = output.getExitNode(c)
        )
      )
    }
  }

  /**
   * `BeegoInputRequestBody` sources of untrusted data.
   */
  private class BeegoInputRequestBodySource extends RemoteFlowSource::Range {
    BeegoInputRequestBodySource() {
      exists(DataFlow::FieldReadNode frn | this = frn |
        frn.getField().hasQualifiedName(contextPackagePath(), "BeegoInput", "RequestBody")
      )
    }
  }

  /**
   * `beego/context.Context` sources of untrusted data.
   */
  private class BeegoContextSource extends RemoteFlowSource::Range {
    BeegoContextSource() {
      exists(Method m | m.hasQualifiedName(contextPackagePath(), "Context", "GetCookie") |
        this = m.getACall().getResult()
      )
    }
  }

  private class BeegoOutputInstance extends Http::ResponseWriter::Range {
    SsaWithFields v;

    BeegoOutputInstance() {
      this = v.getBaseVariable().getSourceVariable() and
      v.getType().(PointerType).getBaseType().hasQualifiedName(contextPackagePath(), "BeegoOutput")
    }

    override DataFlow::Node getANode() { result = v.similar().getAUse().getASuccessor*() }

    /** Gets a header object that corresponds to this HTTP response. */
    DataFlow::MethodCallNode getAHeaderObject() {
      result.getTarget().getName() = ["ContentType", "Header"] and
      this.getANode() = result.getReceiver()
    }
  }

  private class BeegoHeaderWrite extends Http::HeaderWrite::Range, DataFlow::MethodCallNode {
    string methodName;

    BeegoHeaderWrite() {
      this.getTarget().hasQualifiedName(contextPackagePath(), "BeegoOutput", methodName) and
      methodName in ["ContentType", "Header"]
    }

    override DataFlow::Node getName() { methodName = "Header" and result = this.getArgument(0) }

    override string getHeaderName() {
      result = Http::HeaderWrite::Range.super.getHeaderName()
      or
      methodName = "ContentType" and result = "content-type"
    }

    override DataFlow::Node getValue() {
      if methodName = "ContentType"
      then result = this.getArgument(0)
      else result = this.getArgument(1)
    }

    override Http::ResponseWriter getResponseWriter() {
      result.(BeegoOutputInstance).getAHeaderObject() = this
    }
  }

  private class BeegoResponseBody extends Http::ResponseBody::Range {
    DataFlow::MethodCallNode call;
    string methodName;

    BeegoResponseBody() {
      exists(Method m | m.hasQualifiedName(contextPackagePath(), "BeegoOutput", methodName) |
        call = m.getACall() and
        this = call.getArgument(0)
      ) and
      methodName in ["Body", "JSON", "JSONP", "ServeFormatted", "XML", "YAML"]
    }

    override Http::ResponseWriter getResponseWriter() { result.getANode() = call.getReceiver() }

    override string getAContentType() {
      // Super-method provides content-types for `Body`, which requires us to search
      // for `ContentType` and `Header` calls against the same `BeegoOutput` instance
      result = super.getAContentType()
      or
      // Specifically describe methods that set the content-type and body in one operation:
      result = "application/json" and methodName = "JSON"
      or
      result = "application/javascript" and methodName = "JSONP"
      or
      // Actually ServeFormatted can serve JSON, XML or YAML depending on the incoming
      // `Accept` header, but the important bit is this method cannot serve text/html.
      result = "application/json" and methodName = "ServeFormatted"
      or
      result = "text/xml" and methodName = "XML"
      or
      result = "application/x-yaml" and methodName = "YAML"
    }
  }

  private class ControllerResponseBody extends Http::ResponseBody::Range {
    string name;

    ControllerResponseBody() {
      exists(Method m | m.hasQualifiedName(packagePath(), "Controller", name) |
        name = "CustomAbort" and this = m.getACall().getArgument(1)
        or
        name = "SetData" and this = m.getACall().getArgument(0)
      )
    }

    override Http::ResponseWriter getResponseWriter() { none() }

    override string getAContentType() {
      // Actually SetData can serve JSON, XML or YAML depending on the incoming
      // `Accept` header, but the important bit is this method cannot serve text/html.
      result = "application/json" and name = "SetData"
      // CustomAbort doesn't specify a content type, so we assume anything could happen.
    }
  }

  private class ContextResponseBody extends Http::ResponseBody::Range {
    ContextResponseBody() {
      exists(Method m, string name | m.hasQualifiedName(contextPackagePath(), "Context", name) |
        name = "Abort" and this = m.getACall().getArgument(1)
        or
        name = "WriteString" and this = m.getACall().getArgument(0)
      )
    }

    override Http::ResponseWriter getResponseWriter() { none() }

    // Neither method is likely to be used with well-typed data such as JSON output,
    // because there are better methods to do this. Assume the Content-Type could
    // be anything.
    override string getAContentType() { none() }
  }

  private string getALogFunctionName() {
    result =
      [
        "Alert", "Critical", "Debug", "Emergency", "Error", "Info", "Informational", "Notice",
        "Trace", "Warn", "Warning"
      ]
  }

  private class ToplevelBeegoLoggers extends LoggerCall::Range, DataFlow::CallNode {
    ToplevelBeegoLoggers() {
      this.getTarget().hasQualifiedName([packagePath(), logsPackagePath()], getALogFunctionName())
    }

    override DataFlow::Node getAMessageComponent() { result = this.getASyntacticArgument() }
  }

  private class BeegoLoggerMethods extends LoggerCall::Range, DataFlow::MethodCallNode {
    BeegoLoggerMethods() {
      this.getTarget().hasQualifiedName(logsPackagePath(), "BeeLogger", getALogFunctionName())
    }

    override DataFlow::Node getAMessageComponent() { result = this.getASyntacticArgument() }
  }

  private class UtilLoggers extends LoggerCall::Range, DataFlow::CallNode {
    UtilLoggers() { this.getTarget().hasQualifiedName(utilsPackagePath(), "Display") }

    override DataFlow::Node getAMessageComponent() { result = this.getASyntacticArgument() }
  }

  private class HtmlQuoteSanitizer extends SharedXss::Sanitizer {
    HtmlQuoteSanitizer() {
      exists(DataFlow::CallNode c | c.getTarget().hasQualifiedName(packagePath(), "Htmlquote") |
        this = c.getArgument(0)
      )
    }
  }

  /**
   * The File system access sinks
   */
  private class FsOperations extends FileSystemAccess::Range, DataFlow::CallNode {
    int pathArg;

    FsOperations() {
      this.getTarget().hasQualifiedName(packagePath(), "Walk") and pathArg = 1
      or
      exists(Method m | this = m.getACall() |
        m.hasQualifiedName(packagePath(), "FileSystem", "Open") and pathArg = 0
        or
        m.hasQualifiedName(packagePath(), "Controller", "SaveToFile") and pathArg = 1
        or
        m.hasQualifiedName(contextPackagePath(), "BeegoOutput", "Download") and
        pathArg = 0
        or
        // SaveToFileWithBuffer only available in v2
        m.hasQualifiedName("github.com/beego/beego/v2/server/web", "Controller",
          "SaveToFileWithBuffer") and
        pathArg = 1
      )
    }

    override DataFlow::Node getAPathArgument() { result = this.getArgument(pathArg) }
  }

  private class RedirectMethods extends Http::Redirect::Range, DataFlow::CallNode {
    string className;

    RedirectMethods() {
      exists(string package |
        (
          package = packagePath() and className = "Controller"
          or
          package = contextPackagePath() and className = "Context"
        ) and
        this = any(Method m | m.hasQualifiedName(package, className, "Redirect")).getACall()
      )
    }

    override DataFlow::Node getUrl() {
      className = "Controller" and result = this.getArgument(0)
      or
      className = "Context" and result = this.getArgument(1)
    }

    override Http::ResponseWriter getResponseWriter() { none() }
  }

  private class UtilsTaintPropagators extends TaintTracking::FunctionModel {
    UtilsTaintPropagators() { this.hasQualifiedName(utilsPackagePath(), "GetDisplayString") }

    override predicate hasTaintFlow(FunctionInput input, FunctionOutput output) {
      input.isParameter(_) and
      output.isResult(0)
    }
  }
}
